//    DDS; 
//     .
//   STM32F100RB (24).
// : DAC  1 (PA4), DMA1  3, TIM6.

// F0  90    ;
//  250      .

// DAC  1 (PA4)  DMA1  3.
#include <stm32f10x.h>

#include <math.h>

//   ,    
//     Vref/65536.
const int out_max=0xFFFF;
int out_a0=out_max/2;
int out_a1=out_max/3;

//      TIM6.
#define FSYSCLK 24000000
//     ( 
//    ).
//     FSYSCLK/K, 
// K-   2..65536.
//       ,
//    .
#define F0 80000

//    .
unsigned int ph_ind=0;

//       ;
//    f=F0*ph_m/(2.0**32),
//  F0 -  .
unsigned int ph_m=0;

//   .

//      
// (    )    
// : 16    16   .
#define float_to_fix16(x) (int)((x)*0x10000)

//      (16    )
//     .
#define fix16_to_float(x) ((x)/65536.0)

//      (16    ).
//         
//   .
typedef int fix16;

//    sin, |x|<=M_PI/4.
fix16 _fast_sin(fix16 x)
{
	//  x*x: ,  , .
	fix16 x2=(x>>1)*x>>15; // x2=x**2.

	fix16 xn=x2*x>>16;     // xn=x**3.

	fix16 r=x-(xn*float_to_fix16(1/6.0)>>16);

	xn*=x2; xn>>=16;       // xn=x**5.

	// r=x-(x**3)/6.0+(x**5)/120.0.
	r+=(xn*float_to_fix16(1/120.0)>>16);

	return r;
}

//    cos, |x|<=M_PI/4.
fix16 _fast_cos(fix16 x)
{
	//  x*x: ,  , .
	fix16 x2=x*(x>>1)>>15; // x2=x**2.

	fix16 xn=x2*x2>>16;    // xn=x**4.

	fix16 r=float_to_fix16(1.0)-
		(x2*float_to_fix16(0.5)>>16)+
		(xn*float_to_fix16(1/24.0)>>16);

	xn*=x2; xn>>=16;       // xn=x**6.

	// r=1-(x**2)/2.0+(x**4)/24.0-(x**6)/720.0.
	r-=xn*float_to_fix16(1.0/720)>>16;
	
	return r;
}

//   sin  ,   .
fix16 fast_sin(unsigned int k)
{
	fix16 (*ftable[])(fix16) ={_fast_sin, _fast_cos};

	//   k   () phi=2*M_PI*k/(2**32).
	// phi    :
	// phi=n*M_PI/2+alpha,  n -  ,  |alpha|<=M_PI/4.

	//  n.
	unsigned int n=(k+0x20000000)>>30;

	//   alpha   (  ).
	int ka=k-(n<<30);  // fix32, 30bits

	//        fix16.
	fix16 alpha=               // fix30, 31bits
		(int)(4096*2*M_PI)*    // 2*M_PI, fix12, 15bits
		(ka>>14);              // fix18, 16bits
	alpha>>=14;                // fix16

	// r=_fast_sin(alpha)  r=_fast_cos(alpha),  
	//     n:
	fix16 r=ftable[n&1](alpha);

	//   [1]  n ,    
	// r   : if(n&2) r=-r;
	int s=(n>>1)&1;
	return (r^(-s))+s;
}

//  ,   DAC   DMA.
const unsigned int dma_buf_length=64;
short int dma_buf[dma_buf_length];

//    .
//    ph_ind,  ph_m.

//    .
inline void dma_buf_fragment_init(short int *begin, const short int *end)
{
    int a1=out_a1>>1; //    .
    while(begin!=end)
    {
        *begin=out_a0+
            (a1*fast_sin(ph_ind)>>15); //   .
        ph_ind+=ph_m;
        begin++;
    }
}

//      .
inline void dma_buf_first_half_init()
{
    dma_buf_fragment_init(dma_buf, dma_buf+dma_buf_length/2);
}

//      .
inline void dma_buf_second_half_init()
{
    dma_buf_fragment_init(dma_buf+dma_buf_length/2, dma_buf+dma_buf_length);
}

//    DMA ( 3).
extern "C" void DMA1_Channel3_IRQHandler()
{
    if(DMA1->ISR&DMA_ISR_HTIF3)
        dma_buf_first_half_init();
    else if(DMA1->ISR&DMA_ISR_TCIF3)
        dma_buf_second_half_init();
    DMA1->IFCR|=                //  .   1.
            DMA_IFCR_CTEIF3|
            DMA_IFCR_CHTIF3|
            DMA_IFCR_CTCIF3|
            DMA_IFCR_CGIF3;
}

int main()
{
    //       DMA.
    NVIC_SetPriority(DMA1_Channel3_IRQn, 0);
    NVIC_EnableIRQ(DMA1_Channel3_IRQn);

    //     :
    // DMA, DAC, TIM6, GPIOA.
    RCC->AHBENR|=RCC_AHBENR_DMA1EN;
    RCC->APB1ENR|=RCC_APB1ENR_DACEN|RCC_APB1ENR_TIM6EN;
    RCC->APB2ENR|=RCC_APB2ENR_IOPAEN;

    // PA4 -       DAC.
    GPIOA->CRL&=~(0xf<<4*4);

    //       
    //  DAC  DMA.
    DMA1_Channel3->CPAR=(uint32_t)&DAC->DHR12L1; // 
    DMA1_Channel3->CMAR=(uint32_t)dma_buf;       //    .
    DMA1_Channel3->CNDTR=dma_buf_length;         // - .

    uint32_t tmp=DMA1_Channel3->CCR&~0x7FFF|
            DMA_CCR3_TCIE|      //    .
            DMA_CCR3_HTIE|      //   .
            DMA_CCR3_DIR|       // : ->.
            DMA_CCR3_CIRC|      //   .
            // DMA_CCR3_PINC=0  //    .
            DMA_CCR3_MINC|      //    .
            DMA_CCR3_PSIZE_1|   //   =10: 32 .
            DMA_CCR3_MSIZE_0|   //    =01: 16 .
            DMA_CCR3_PL;        // 11 -  .
    DMA1_Channel3->CCR=tmp;
    DMA1_Channel3->CCR|=DMA_CCR3_EN;

    //    DAC.
    DAC->CR|=
        DAC_CR_DMAEN1| //    DMA.
        DAC_CR_TEN1;   //   DAC    .
        // TSEL[2:0]=000:   Timer 6 TRGO event.
    DAC->CR|=DAC_CR_EN1;

    //       -
    //  TIM6.     
    //    f=F0*ph_m/(2.0**32) =>
    // ph_m=(uint32_t)((2.0**32)*f/F0);

    //  (  F0=80):
    ph_m=2684354; // 50 
    //ph_m=2689723; // 50.1 
    //ph_m=268435456; // 5000 
    //ph_m=268435456*2; // 10000 
    //ph_m=662803111; // 12345.67Hz

    //        DAC,
    //         DMA.
    dma_buf_first_half_init();
    dma_buf_second_half_init();

    TIM6->ARR=FSYSCLK/F0-1;   //  : FSYSCLK/(ARR+1)=F0.
    TIM6->CR2&=~TIM_CR2_MMS;
    TIM6->CR2|=TIM_CR2_MMS_1; //MMS[2:0]=010:  TRGO  UEV.
    TIM6->CR1|=TIM_CR1_CEN;   //  .

    while(true);
}
